#include "Predictor.h"
#include "opencv2/opencv.hpp"
#include <algorithm>
using namespace std;
using namespace cv;
namespace autocar
{
namespace vision_mul
{

void Predictor::setFrames()
{
  
    if(!history_value.empty())
    {
      for(auto i:history_value)
      {
	i = i + old_first_frame - first_frame;
      }
    }
}

bool Predictor::setRecord(double value, double time){
    if(!history_value.empty())
    {
      if (history_value.size() < history_size){
	  old_first_frame = history_time.front();
	  history_time.push_back(time);
	  history_value.push_back(value);
	  first_frame = history_time.front();
      }
      else {
	  old_first_frame = history_time.front();
	  history_time.push_back(time);
	  history_value.push_back(value);
	  history_time.pop_front();
	  history_value.pop_front();
	  first_frame = history_time.front();
      }
    }
    else
    {
	old_first_frame = 0;
	history_time.push_back(time);
	history_value.push_back(value);
	first_frame = history_time.front();
    }   
}

double Predictor::predict(double time){
    
    list<double>::const_iterator it_in = history_time.begin();
    double latest_value = history_value.back();
    if(abs(latest_value) < (f * 5.0 /57.3) || history_value.size() < history_size)
        return latest_value;//上次记录值 小于5
 
    if(history_time.back() - *it_in > 150.0)
        return latest_value;//上次记录值 与第一帧值相差 150针
    list<double>::const_iterator it_out = history_value.begin();
    list<double>::const_iterator prev_out = it_out;
    double max_o = -500000, min_o = 500000;
    for(list<double>::const_iterator it = it_out, it_i = it_in; it != history_value.end(); ++it, ++it_i){
        if(max_o < *it)
            max_o = *it;
        if(min_o > *it)
            min_o = *it;
        if(abs(*it - *prev_out) > (f * 5.0  / 57.3 )){
            return latest_value;
        }
        prev_out = it;
        //printf("(%2f,%2f) ", *it, *it_i);
    }
   // printf("\n"); 
    if (max_o - min_o < (f * 3.0 / 57.3 ))      // angle gap must lager than 3 degree
        return latest_value;

    Mat A(history_size,3,CV_64F);
    Mat b(history_size,1,CV_64F);
    double * b_data = (double *) b.data;
    double * A_data = (double *) A.data;

    for (; it_in != history_time.end(); ++A_data, ++b_data, ++it_in, ++it_out){
        *A_data = (*it_in-time) * (*it_in-time);
        *(++A_data) = (*it_in-time);
        *(++A_data) = 1;
        *b_data = *it_out;
    }

    // A*w = b  =>  w = (A_t * A).inverse * b
    Mat A_t = A.t();
    Mat w = (A_t*A).inv()*A_t * b;
    Mat q = (Mat_<double>(1,3) << 0, 0, 1);
    Mat ret = q*w;

    double predict_angel = ret.at<double>(0);
    const double max_gap =( f * 3.0 / 57.3  ) ;
    if(predict_angel - latest_value > max_gap)
        predict_angel = latest_value + max_gap;
    else if(predict_angel - latest_value < -max_gap)
        predict_angel = latest_value - max_gap;
    return predict_angel;

}	
}
}

