#include "armor_detect_node.h"
#include "debug_utility.hpp"
#include "util.h"
#include "draw.h"
#include "opencv2/opencv.hpp"
#include "vision_unit/armor_msg.h"
#include "labeler.h"
#include "video_recoder.h"
#include <memory>
#include <math.h>
#include <cv_bridge/cv_bridge.h>
#include "Settings.h"
#include "AngleSolver.h"
#ifndef _DEBUG_VISION
#define _DEBUG_VISION
#endif
#ifndef PREDICT_XY
#define PREDICT_XY
#endif

namespace autocar
{
namespace vision_mul
{

  
armor_detect_node::armor_detect_node(ros::NodeHandle& nh,ros::NodeHandle& pnh):it_(nh)
{
   // detector= boost::shared_ptr<ArmorDetector>(new ArmorDetector());
    image_sub_ = it_.subscribeCamera("/usb_cam/image_raw", 1, &armor_detect_node::detection, this);
    sub_yaw = nh.subscribe<serial_comm::car_speed>("car_info", 5, boost::bind(&armor_detect_node::pan_info_callback, this, _1));
    pub_armor_pos = nh.advertise<vision_unit::armor_msg>("armor_info", 1000);
    pub_goal = nh.advertise<move_base_msgs::MoveBaseGoal>("camera_goal", 10);
    pnh.param<bool>("/armor_detect/debug_on", debug_on, false);
    std::cout<<"Debug_on:"<<debug_on<<std::endl;

    svm = cv::ml::SVM::load("/home/wade/Desktop/RMVision/SVM=====.xml");
    //Set the goal value to geometry_msgs::PoseStamped.
    goal_pose.target_pose.header.stamp    = ros::Time::now();
    goal_pose.target_pose.header.frame_id = "base_link";
    goal_pose.target_pose.pose.position.x = 0;
    goal_pose.target_pose.pose.position.y = 0;
    goal_pose.target_pose.pose.position.z = 0;


    goal_pose.target_pose.pose.orientation.w = 0;
    goal_pose.target_pose.pose.orientation.x = 0;
    goal_pose.target_pose.pose.orientation.y = 0;

    //init pnp param
    obj_p.push_back(cv::Point3f(0,   0,  0.0));
    obj_p.push_back(cv::Point3f(124,   0,  0.0));
    obj_p.push_back(cv::Point3f(124,   54,   0.0));
    obj_p.push_back(cv::Point3f(0,    54,   0.0));

    camera_matrix = cv::Mat(3, 3, CV_64F, cam);
    dist_coeffs   = cv::Mat(5, 1, CV_64F, dist_c);

    rvec = cv::Mat::ones(3,1,CV_64F);
    tvec = cv::Mat::ones(3,1,CV_64F);
    forward_back = true;
  //  detect_start_time = std::chrono::system_clock::now();
  //  cam_time = detect_start_time;
    start_time = ros::Time::now();
    current_time = start_time;
    cout<<"construction finished"<<endl;
    //cv::Mat(obj_p).convertTo(obj_points, CV_32F);
}

armor_detect_node::~armor_detect_node(void)
{
    image_sub_.shutdown();
}

void armor_detect_node::detection(const sensor_msgs::ImageConstPtr& msg,const sensor_msgs::CameraInfoConstPtr& cam_info)
{
    int t1 = cv::getTickCount();
    cv::Mat template_img = cv::imread("/home/wade/Desktop/RMVision/template.bmp");
    cv::Mat small_template_img = cv::imread("/home/wade/Desktop/RMVision/small_template.bmp");
    char * config_file_name = "/home/wade/Desktop/RMVision/param_config.xml";
    Settings setting(config_file_name);
    OtherParam other_param;
    FileStorage fs(setting.intrinsic_file_480, FileStorage::READ);
    if (!fs.isOpened())	{
        std::cout << "Could not open the configuration file: \"" << setting.intrinsic_file_480 << "\"" << std::endl;
        return ;
    }
    Mat cam_matrix_480, distortion_coeff_480;
    fs["Camera_Matrix"] >> cam_matrix_480;
    fs["Distortion_Coefficients"] >> distortion_coeff_480;
    
    FileStorage fs1(setting.intrinsic_file_720, FileStorage::READ);
    if (!fs1.isOpened())	{
        cout << "Could not open the configuration file: \"" << setting.intrinsic_file_720 << "\"" << endl;
        return ;
    }
    Mat cam_matrix_720, distortion_coeff_720;
    fs1["Camera_Matrix"] >> cam_matrix_720;
    fs1["Distortion_Coefficients"] >> distortion_coeff_720;
    AngleSolver solver_480(cam_matrix_480, distortion_coeff_480, 12.4, 5.4, setting.scale_z_480, setting.min_detect_distance, setting.max_detect_distance);
    AngleSolver solver_720(cam_matrix_720, distortion_coeff_720, 12.4, 5.4, setting.scale_z, setting.min_detect_distance, setting.max_detect_distance);
    
    Point2f image_center_480 = cv::Point2f(cam_matrix_480.at<double>(0,2), cam_matrix_480.at<double>(1,2));
    Point2f image_center_720 = cv::Point2f(cam_matrix_720.at<double>(0,2), cam_matrix_720.at<double>(1,2));
    
    
    
    //coordinate transform
    const double overlap_dist = 100000.0;
    const double barrel_ptz_offset_y = 3.3;//炮口 炮管相对云台坐标系的y偏差
    const double ptz_camera_y = 0.8;//云台坐标系 相对相机坐标系y轴的偏移
    const double ptz_camera_z = -20.5;//云台坐标系 相对相机坐标系z轴的偏移
    //const double ptz_camera_z = -10.5;
    double theta = -atan((ptz_camera_y + barrel_ptz_offset_y)/overlap_dist);//
    double r_data[] = {1,0,0,0,cos(theta),-sin(theta), 0, sin(theta), cos(theta)};
    double t_data[] = {0, ptz_camera_y, ptz_camera_z}; // ptz org position in camera coodinate system
    Mat t_camera_ptz(3,1, CV_64FC1, t_data);
    Mat r_camera_ptz(3,3, CV_64FC1, r_data);
    solver_480.setRelationPoseCameraPTZ(r_camera_ptz, t_camera_ptz, barrel_ptz_offset_y);
    solver_720.setRelationPoseCameraPTZ(r_camera_ptz, t_camera_ptz, barrel_ptz_offset_y);
    
    
    AngleSolverFactory angle_slover;
    angle_slover.setTargetSize(21.6, 5.4, AngleSolverFactory::TARGET_ARMOR);
    angle_slover.setTargetSize(12.4, 5.4, AngleSolverFactory::TARGET_SAMLL_ATMOR);
    angle_slover.setTargetSize(28.0, 16.0, AngleSolverFactory::TARGET_RUNE);
    
    detector.initTemplate(template_img,small_template_img);
    detector.setPnPSlover(&solver_480, &solver_720);
    detector.setPara(setting.armor);
    angle_slover.setSolver(&solver_720);
    detector.setPitchAngle(other_param.angle_pitch);

    
    
    cv_bridge::CvImagePtr cv_ptr;
    int64_t currentTime = 0;
    x_predictor.setFocus(cam[0]);
    y_predictor.setFocus(cam[4]);
    try{
      cv_ptr = cv_bridge::toCvCopy(msg, sensor_msgs::image_encodings::BGR8);
      nframes++;
      //cam_time = std::chrono::system_clock::now();
     // current_time = ros::Time::now();
      //ros::Duration d= current_time - start_time;
   //   currentTime = current_time.toNSec();
     // cout<<"Current Time is:"<<currentTime<<endl;
     }
     catch (cv_bridge::Exception& e){
      ROS_ERROR("cv_bridge exception: %s", e.what());
      return;
    }
     cv::Mat image = cv_ptr->image.clone();
     std::cout<<"Show image"<<std::endl;
    //int nframes = 0;
    
    bool detected = false;
    //filter
    int detected_count = 0;
    int undetected_count = 3;
    
   // RotatedRect rect;
  // for (;;)
    
      
     //  std::chrono::duration<double> current_time;
    //   std::chrono::duration<double> diff = cam_time-detect_start_time;
     //  double times = diff.count();
      // cout<"current Time: "<<times<<endl;
        //if(forward_back)
        //capture_camera_forward >> image;
        if(image.empty())
        {
            std::cout<<"Image has no data!"<<std::endl;
	    return ;
            //continue;
        }
        //else
        //capture_camera_back >> image;

        // Find the fianl armor.
        if(forward_back && !image.empty())
        {
	  detector.setsvm(svm);
	  detected =  detector.detect(image);
	   // rect = armor_detector.getTargetAera(image);
           // detected = armor_detector->detect(image, true);
        }
        if(nframes > LONG_MAX /2 )
	{
	      nframes = 0;
	      x_predictor.clear();
	      y_predictor.clear();
	}
        if(detected)
        {
            detected_count++;
            armor_2 = detector.get_armor();
            if(detected_count >= 1)
            {
                detected_armor = true;
		x_predictor.setRecord(armor_2.armor.center.x,nframes);
		x_predictor.setFrames();
	        y_predictor.setRecord(armor_2.armor.center.y,nframes);
		y_predictor.setFrames();
                detected_count = 0;
		//std::cout<<"armor detected#############"<<std::endl;
            }
        }
        else
        {
            if(undetected_count != 0)
                undetected_count--;
            if(undetected_count == 0)
            {
                detected_armor = false;
                undetected_count = 25;
            }
        }

#ifdef _DEBUG_VISION
      //  int keycode = cv::waitKey(1);
    //    if ((keycode & 0xff) == 27) // ESC
     //       break;
#endif

        //forward_back = !forward_back;
        //ros
        //forward_back = get_camera_num();
        if(!forward_back)
        {
            if(detected_armor)
            {
                goal_pose.target_pose.pose.orientation.z = 1;//false
                //goal.request.target_pose = goal_pose;
                pub_goal.publish(goal_pose);
            }
        }
        else
        {
            if(detected_armor&&(!armor_2.points.empty()))
            {
                //set the image_points
                //        std::cout<<"point size: "<<armor_->points.size()<<std::endl;
                //        std::cout<<"point 1: "<<armor_->points[0]<<std::endl;
                //        std::cout<<"point 2: "<<armor_->points[1]<<std::endl;
                //        std::cout<<"point 3: "<<armor_->points[2]<<std::endl;
                //        std::cout<<"point 4: "<<armor_->points[3]<<std::endl;
                //set_image_points(armor_->points);

                cv::solvePnP(obj_p, armor_2.points, camera_matrix, dist_coeffs, rvec, tvec);
                double enemy_dis = tvec.at<double>(2)/1000.0;//Z distance
                double enemy_ang = rvec.at<double>(1);//Rotate around y axis
		double x_pre,y_pre;
                for(int i =0 ; i < armor_2.points.size(); i++)
                {
                	std::cout<<"Armor Rect Point "<<i<<" is: "<<armor_2.points[i].x -640 <<" "<<armor_2.points[i].y -360<<std::endl;
                }
                std::cout<<"Enemy distance:"<<enemy_dis*1000<<" mm"<<std::endl;
                //std::cout<<"Enemy angle:"<<enemy_ang*180/3.14<<std::endl;
                if(enemy_dis < 4.5)
                {
		    armor_pos.detected = true;
#ifdef PREDICT_XY
		    x_pre = x_predictor.predict(nframes + 0.5);
		    //y_pre = y_predictor.predict(nframes + 1.0);
		    y_pre = armor_2.armor.center.y;
		    
#else
		    x_pre = armor_2.armor.center.x;
		    y_pre = armor_2.armor.center.y;
#endif
                    //armor_pos.detected = true;
                    armor_pos.x = x_pre - cam[2];//转换成角度信息 f
                    armor_pos.y = y_pre - cam[5];
                    armor_pos.d = enemy_dis*1000;
		    
                }
                else
		{
		   armor_pos.detected = false;
		}
#ifdef _DEBUG_VISION
      //  int keycode = cv::waitKey(1);
    //    if ((keycode & 0xff) == 27) // ESC
     //       break;
		cv::Point2f  image_center = Point2f(640,360);
		cv::Point2f  predict_pos = Point2f(x_pre,y_pre);
		circle(image, image_center, 3, CV_RGB(0, 255, 0), 2);
		circle(image, predict_pos,5,CV_RGB(255, 0, 0), 3);
		Point2f vertices[4];
		armor_2.armor.points(vertices);
		for (int i = 0; i < 4; i++){
		    line(image, vertices[i], vertices[(i + 1) % 4], CV_RGB(0, 255, 0), 3);
		}
		char str[30];
		sprintf(str, "%.1f, %.1f, %f", armor_2.armor.center.x-640, armor_2.armor.center.y-360, enemy_dis);
		cv::putText(image, str, Point(10, 40), CV_FONT_HERSHEY_COMPLEX_SMALL, 1.3, CV_RGB(128, 255, 0), 1);
		
#endif

		
                //Publishing armor data tp serial_write
                pub_armor_pos.publish(armor_pos);


                //double goal_dis = enemy_dis - 2.0*sin(pan_yaw);
                //double goal_dis = enemy_dis - 2.5;
                goal_pose.target_pose.header.frame_id = "base_link";
                goal_pose.target_pose.header.stamp = ros::Time::now();
                goal_pose.target_pose.pose.position.x = enemy_dis;
                goal_pose.target_pose.pose.position.y = 0;
                goal_pose.target_pose.pose.position.z = 1;
                //tf::Quaternion q;
                //q.setRPY(0, 0, pan_yaw);
                goal_pose.target_pose.pose.orientation.x= 0;
                goal_pose.target_pose.pose.orientation.y= 0;
                goal_pose.target_pose.pose.orientation.z= 0;
                goal_pose.target_pose.pose.orientation.w= 1;
                pub_goal.publish(goal_pose);
                armor_2.points.clear();
            }
            else
            {
                armor_pos.detected = false;
                pub_armor_pos.publish(armor_pos);

                goal_pose.target_pose.pose.position.z = 0;
                pub_goal.publish(goal_pose);
            }
        }
	int t2 = cv::getTickCount();
	std::cout << "Consumer-Time: " << (t2 - t1) * 1000.0 / cv::getTickFrequency() << "ms   "  <<std::endl;
	imshow("result", image);
	waitKey(1);
}
void armor_detect_node::running(void)
{
    cv::Mat image;
    //cv::VideoCapture capture_camera_forward("/data/dl_armor_detection/dataset/armor_0002.avi");
    cv::VideoCapture capture_camera_forward(0);
    if(!capture_camera_forward.isOpened())
    {
        std::cout<<"Cannot open the camera!"<<std::endl;
        return;
    }
    set_camera_exposure("/dev/video1", 65);
    //{
    //  while (true) {
    //    boost::this_thread::sleep_for(boost::chrono::seconds(10));
    //  }
    //}

    cv::VideoCapture capture_camera_back("/home/dji/Videos/armor_20170405_170846.avi");
    std::shared_ptr<armor_detecter> armor_detector;
    std::shared_ptr<labeler> label;
    std::shared_ptr<video_recoder> recoder;

    int nframes = 0;
    bool detected = false;
    //filter
    int detected_count = 0;
    int undetected_count = 1;

    for (;;)
    {
        auto speed_test_start_begin_time = std::chrono::system_clock::now();
        //if(forward_back)
        capture_camera_forward >> image;
        if(image.empty())
        {
            std::cout<<"Image has no data!"<<std::endl;
            continue;
        }
        //else
        //capture_camera_back >> image;

        // Find the fianl armor.
        if (armor_detector == nullptr)
            armor_detector = std::shared_ptr<armor_detecter>(new armor_detecter(debug_on));
        if(forward_back && !image.empty())
        {
            detected = armor_detector->detect(image, true);
        }

        if(detected)
        {
            detected_count++;
            armor_ = armor_detector->get_armor();
            if(detected_count >= 1)
            {
                detected_armor = true;
                detected_count = 0;
            }
        }
        else
        {
            if(undetected_count != 0)
                undetected_count--;
            if(undetected_count == 0)
            {
                detected_armor = false;
                undetected_count = 1;
            }
        }

#ifdef _DEBUG_VISION
        int keycode = cv::waitKey(1);
        if ((keycode & 0xff) == 27) // ESC
            break;
#endif

        //forward_back = !forward_back;
        //ros
        //forward_back = get_camera_num();
        if(!forward_back)
        {
            if(detected_armor)
            {
                goal_pose.target_pose.pose.orientation.z = 1;
                //goal.request.target_pose = goal_pose;
                pub_goal.publish(goal_pose);
            }
        }
        else
        {
            if(detected_armor)
            {
                //set the image_points
                //        std::cout<<"point size: "<<armor_->points.size()<<std::endl;
                //        std::cout<<"point 1: "<<armor_->points[0]<<std::endl;
                //        std::cout<<"point 2: "<<armor_->points[1]<<std::endl;
                //        std::cout<<"point 3: "<<armor_->points[2]<<std::endl;
                //        std::cout<<"point 4: "<<armor_->points[3]<<std::endl;
                //set_image_points(armor_->points);

                cv::solvePnP(obj_p, armor_->points, camera_matrix, dist_coeffs, rvec, tvec);
               if(!armor_->points.empty()) 
               {
                for(int i =0 ; i < armor_->points.size(); i++)
                {
                	std::cout<<"Armor Rect Point "<<i<<" is: "<<armor_->points[i].x<<" "<<armor_->points[i].y<<std::endl;
                }
               }
                double enemy_dis = tvec.at<double>(2)/1000.0;//Z distance
                double enemy_ang = rvec.at<double>(1);//Rotate around y axis
                std::cout<<"Enemy distance:"<<enemy_dis<<" m"<<std::endl;
                //std::cout<<"Enemy angle:"<<enemy_ang*180/3.14<<std::endl;
                if(enemy_dis < 4.5)
                {
                    armor_pos.detected = true;
                    armor_pos.x = armor_->armor.center.x - 320;//转换成角度信息 f
                    armor_pos.y = armor_->armor.center.y - 240;
                    armor_pos.d = enemy_dis*1000;
                }
                else
                    armor_pos.detected = false;
                //Publishing armor data tp serial_write
                pub_armor_pos.publish(armor_pos);


                //double goal_dis = enemy_dis - 2.0*sin(pan_yaw);
                //double goal_dis = enemy_dis - 2.5;
                goal_pose.target_pose.header.frame_id = "base_link";
                goal_pose.target_pose.header.stamp = ros::Time::now();
                goal_pose.target_pose.pose.position.x = enemy_dis;
                goal_pose.target_pose.pose.position.y = 0;
                goal_pose.target_pose.pose.position.z = 1;
                //tf::Quaternion q;
                //q.setRPY(0, 0, pan_yaw);
                goal_pose.target_pose.pose.orientation.x= 0;
                goal_pose.target_pose.pose.orientation.y= 0;
                goal_pose.target_pose.pose.orientation.z= 0;
                goal_pose.target_pose.pose.orientation.w= 1;
                pub_goal.publish(goal_pose);
            }
            else
            {
                armor_pos.detected = false;
                pub_armor_pos.publish(armor_pos);

                goal_pose.target_pose.pose.position.z = 0;
                pub_goal.publish(goal_pose);
            }
        }
    }
}

void armor_detect_node::pan_info_callback(const serial_comm::car_speed::ConstPtr &pan_data)
{
    //pan_yaw = pan_data->yaw;
}

bool armor_detect_node::if_detected_armor()
{
    return detected_armor;
}

bool armor_detect_node::get_camera_num()
{
    return forward_back;
}

armor_info* armor_detect_node::get_armor()
{
    return armor_;
}

void armor_detect_node::set_image_points(std::vector<cv::Point2f> armor_points)
{
    //  cv::Point2f vertices[4];
    //  rect.points(vertices);
    //  cv::Point2f lu, ld, ru, rd;
    //  std::sort(vertices, vertices + 4, [](const cv::Point2f & p1, const cv::Point2f & p2) { return p1.x < p2.x; });
    //  if (vertices[0].y < vertices[1].y){
    //    lu = vertices[0];
    //    ld = vertices[1];
    //  }
    //  else{
    //    lu = vertices[1];
    //    ld = vertices[0];
    //  }
    //  if (vertices[2].y < vertices[3].y)	{
    //    ru = vertices[2];
    //    rd = vertices[3];
    //  }
    //  else {
    //    ru = vertices[3];
    //    rd = vertices[2];
    //  }
    //  //img_p.push_back()
    //  img_p.clear();
    //  img_p.push_back(lu);
    //  img_p.push_back(ru);
    //  img_p.push_back(rd);
    //  img_p.push_back(ld);
    cv::Mat(armor_points).convertTo(img_points, CV_32F);
}
}
}

