#ifndef __THREAD_OPERATION_H__
#define __THREAD_OPERATION_H__
//
#include <ros/ros.h>
#include <image_transport/image_transport.h>
#include<tf/tf.h>
#include <boost/thread.hpp>
#include <boost/atomic.hpp>
#include <opencv2/opencv.hpp>
#include <logical_core/SetGoal.h>
#include <vision_unit/armor_msg.h>
#include <geometry_msgs/PoseStamped.h>
#include <geometry_msgs/PoseWithCovarianceStamped.h>
#include <move_base_msgs/MoveBaseAction.h>
#include <serial_comm/car_speed.h>
#include "armor_detect.h"
#include "armor_detector.h"
#include "Predictor.h"
#include <cmath>
#include <climits>
namespace autocar
{
namespace vision_mul
{
class armor_detect_node
{
public:
    armor_detect_node(ros::NodeHandle& nh,ros::NodeHandle& pnh);
    ~armor_detect_node(void);
     /**
     * @brief running the detection process of armor detector
     */
    void detection(const sensor_msgs::ImageConstPtr& msg,const sensor_msgs::CameraInfoConstPtr& cam_info);
    /**
     * @brief running the program of armor detect
     */
    void running(void);
    /**
     * @brief If the program can detect an armor
     * @return ture if detected an armor
     */
    bool if_detected_armor();
    void pan_info_callback(const serial_comm::car_speed::ConstPtr &pan_data);
    /**
     * @brief Maybe we have more than one camera, so we can use this function to get the camera No. that detected armor
     * @return
     */
    bool get_camera_num();
    /**
     * @brief Obtaining the armor information including the width and hight of the armor
     * @return
     */
    armor_info* get_armor();

private:
    void set_image_points(std::vector<cv::Point2f> armor_points);

private:
    image_transport::ImageTransport it_;
    image_transport::CameraSubscriber image_sub_;
    //vision_param *param;
   // ros::NodeHandle n;
    bool detected_armor;
    armor_info *armor_;
    armor2_info armor_2;
    ArmorDetector detector;
    float pan_yaw;
    //auto detect_start_time;
    //auto cam_time;
    //std::chrono::duration<double> timestamp;
    cv::Mat m_img_buff[2];
    bool forward_back;
    bool debug_on;
    ros::Subscriber sub_yaw;
    ros::Publisher pub_armor_pos, pub_goal;
    vision_unit::armor_msg armor_pos;
    move_base_msgs::MoveBaseGoal goal;
    move_base_msgs::MoveBaseGoal goal_pose;
    ros::Time start_time;
    ros::Time current_time;
    Predictor x_predictor;
    Predictor y_predictor;
    long  nframes ;
   // uint64_t current_time;
    cv::Mat camera_matrix;
    cv::Mat dist_coeffs;
    cv::Mat obj_points;
    cv::Mat img_points;
    std::vector<cv::Point3f> obj_p;
    std::vector<cv::Point2f> img_p;
    cv::Mat rvec;
    cv::Mat tvec;
    double ang_lim = 3.0/180*3.14;
    double cam[9] = {1231.462061656437, 0, 639.7304027046221, 0, 1233.278111018447, 381.3793167827258, 0, 0, 1};

    double dist_c[5] = {-0.4456573334760076, 0.2268579138890603, -5.115779213883301e-05, -6.095460843668778e-05, 0};
};
}
}

#endif // !__THREAD_OPERATION_H__

