#ifndef __ARMOR2_H__
#define __ARMOR2_H__

#include <opencv2/opencv.hpp>
#include <vector>
#include <list>
#include <tuple>
#include <chrono>
#include "AngleSolver.h"
#include "detect_factory.h"
#define TRUNC_ABS(a) ((a) > 0 ? (a) : 0);
#define POINT_DIST(p1,p2) std::sqrt((p1.x-p2.x)*(p1.x-p2.x) + (p1.y-p2.y)*(p1.y-p2.y))
enum EnemyColor { RED = 0, BLUE = 1};

using namespace cv;
using namespace std;
struct ArmorParam {
	uchar min_light_gray;	        // 
	uchar min_light_height;			// 
	uchar avg_contrast_threshold;	// 
	uchar light_slope_offset;		// 
	int  max_light_delta_h;         // 
	uchar min_light_delta_h;		// 
	uchar max_light_delta_v;		// 
	uchar max_light_delta_angle;	// 
	uchar avg_board_gray_threshold; // 
	uchar avg_board_grad_threshold; // 
	uchar grad_threshold;			//  (not used)
	uchar br_threshold;				//
	uchar enemy_color;                 // 0 for red, otherwise blue

    ArmorParam(){
        min_light_gray = 170;
	min_light_height = 5;
        avg_contrast_threshold = 110;
	light_slope_offset = 30;
        max_light_delta_h = 280;
	min_light_delta_h = 16;
	max_light_delta_v = 15;
	max_light_delta_angle = 20;
        avg_board_gray_threshold = 60;
        avg_board_grad_threshold = 25;
        grad_threshold = 20;
	br_threshold = 0;
        enemy_color = 1;
	}
};
class armor2_info
{
public:

  armor2_info(int type_ = 0, cv::RotatedRect armor_=cv::RotatedRect(cv::Point2f(0,0), cv::Size2f(0,0), 0))
  {
    type = type_;
    armor = armor_;
  }
  int type;
  cv::RotatedRect armor;
  std::vector<cv::Point2f> points;
};
namespace autocar
{
namespace vision_mul
{






  class ArmorDetector {
    public:
    ArmorDetector(const ArmorParam & para = ArmorParam()){
        pitch_angle = 0;
        s_solver = NULL;
        l_solver = NULL;
        _para = para;
        _res_last = cv::RotatedRect();
        _dect_rect = cv::Rect();
        _is_small_armor = false;
        _lost_cnt = 0;
        _is_lost = true;
	//armor(0,RotatedRect(Point2f(0,0), Size2f(0,0), 0),)
	}
    void setPara(const ArmorParam & para){
        _para = para;
    }

    void setPnPSlover(AngleSolver * solver_s, AngleSolver * solver_l){
        s_solver = solver_s;
        l_solver = solver_l;
    }

    void setPitchAngle(double angle){
        pitch_angle = angle;
    }

    void reset(){
        _res_last = cv::RotatedRect();
        _dect_rect = cv::Rect();
        _is_small_armor = false;
        _lost_cnt = 0;
        _is_lost = true;
    }

    bool isSamllArmor(){
        return _is_small_armor;
    }

    void initTemplate(const cv::Mat & _template, const cv::Mat & _template_small);
    cv::RotatedRect getTargetAera(const cv::Mat & src);
    void setLastResult(const cv::RotatedRect & rect){
        _res_last = rect;
    }
    const cv::RotatedRect & getLastResult() const{
        return _res_last;
    }
    
    armor2_info get_armor();
   bool detect(const cv::Mat &image);
private:
    /**
     * @brief setImage Pocess the input (set the green component and sub of blue and red component)
     * @param src
     */
    void setImage(const cv::Mat & src);

    /**
     * @brief templateDist Compute distance between template and input image
     * @param img input image
     * @param is_smarect_pnp_solverll true if input image is a samll armor, otherwise, false
     * @return distance
     */
    int templateDist(const cv::Mat & img, bool is_small);

    /**
     * @brief findContourInEnemyColor Find contour in _max_color
     * @param left output left contour image (probably left lamp of armor)
     * @param right output righe edge (probably right lamp of armor)
     * @param contours_left output left contour
     * @param contours_right output right contour
     */
    void findContourInEnemyColor(
        cv::Mat & left, cv::Mat & right,
        std::vector<std::vector<cv::Point2i> > &contours_left,
        std::vector<std::vector<cv::Point2i> > &contours_right);

    /**
     * @brief findTargetInContours Find target rectangles
     * @param contours_left input left contour
     * @param contours_right input right contour
     * @param rects target rectangles (contains wrong area)
     */
    void findTargetInContours(
        const std::vector<std::vector<cv::Point> > & contours_left,
        const std::vector<std::vector<cv::Point> > & contours_right,
        std::vector<cv::RotatedRect> & rects,
        std::vector<double> & score);

    /**
     * @brief chooseTarget Choose the most possible rectangle among all the rectangles
     * @param rects candidate rectangles
     * @return the most likely armor (RotatedRect() returned if no proper one)
     */
    cv::RotatedRect chooseTarget(const std::vector<cv::RotatedRect> & rects, const std::vector<double> & score);

    /**
     * @brief boundingRRect Bounding of two ratate rectangle (minumum area that contacts two inputs)
     * @param left left RotatedRect
     * @param right right RotatedRect
     * @return minumum area that contacts two inputs
     */
    cv::RotatedRect boundingRRect(const cv::RotatedRect & left, const cv::RotatedRect & right);

    /**
     * @brief adjustRRect Adjust input angle
     * @param rect input
     * @return adjusted rotate rectangle
     */
    cv::RotatedRect adjustRRect(const cv::RotatedRect & rect);
    bool makeRectSafe(cv::Rect & rect, cv::Size size){
        if (rect.x < 0)
            rect.x = 0;
        if (rect.x + rect.width > size.width)
            rect.width = size.width - rect.x;
        if (rect.y < 0)
            rect.y = 0;
        if (rect.y + rect.height > size.height)
            rect.height = size.height - rect.y;
        if (rect.width <= 0 || rect.height <= 0)
            return false;
        return true;
    }

    bool broadenRect(cv::Rect & rect, int width_added, int height_added, cv::Size size){
        rect.x -= width_added;
        rect.width += width_added * 2;
        rect.y -= height_added;
        rect.height += height_added * 2;
        return makeRectSafe(rect, size);
    }

    void cal_armor_info(std::vector<cv::Point2f> &armor_points, cv::RotatedRect armor_rect);
    
 
private:
    AngleSolver * s_solver;
    AngleSolver * l_solver;
    double pitch_angle;
    bool _is_lost;
    int _lost_cnt;
    bool _is_small_armor;           // true if armor is the small one, otherwise false
    cv::RotatedRect _res_last;      // last detect result
    cv::Rect _dect_rect;            // detect roi of original image
    ArmorParam _para;               // parameter of alg
    cv::Mat _binary_template;       // armor template binary image
    cv::Mat _binary_template_small; // small armor template binay image
    cv::Mat _src;                   // source image
    cv::Mat _g;                     // green component of source image
    cv::Mat _ec;                    // enemy color
    cv::Mat _max_color;             // binary image of sub between blue and red component
    cv::Size _size;
    cv::Mat res_im;
    armor2_info armor;
    //armor_info* old_armor;
//    cv::Mat _gray;
//    cv::Mat _b;
//    cv::Mat _r;
  };
  
}
}
#endif // !__ARMOR2_H__